/**
 * \file cantileverWingDesignTaw.cpp
 * \author Christopher Ruwisch (christopher.ruwisch@tu-berlin.de)
 * \brief
 * \version 0.1
 * \date 2023-12-05
 *
 * @copyright Copyright (c) 2023
 *
 */

#include "cantileverWingDesignTaw.h"

#include <aircraftGeometry2/io/dat.h>
#include <aircraftGeometry2/processing/measure.h>
#include <aircraftGeometry2/processing/transform.h>
#include <atmosphere/atmosphere.h>
#include <unitConversion/unitConversion.h>

#include <algorithm>
#include <cmath>
#include <map>
#include <tuple>

#include "lib/design_methods/aspectRatioMethods.h"
#include "lib/design_methods/controlDeviceMethods.h"
#include "lib/design_methods/dihedralMethods.h"
#include "lib/design_methods/relativeKinkPositionMethods.h"
#include "lib/design_methods/sparMethods.h"
#include "lib/design_methods/sweepAngleMethods.h"
#include "lib/design_methods/taperRatioMethods.h"
#include "lib/design_methods/wingAreaMethods.h"
#include "lib/design_methods/wingProfileMethods.h"
#include "lib/general_methods/icao_faa_codes.h"
#include "lib/general_methods/newtonAlgorithm.h"

geom2::AirfoilSection get_rescaled_airfoil_geometry(const std::string& airfoildatadir,
                                                    std::tuple<double, double, std::string> wing_profile_and_thickness);

namespace taw {
namespace cantilever {
Wing::Wing(const std::shared_ptr<RuntimeIO>& rtIO)
    : rtIO(rtIO),
      config(std::make_shared<taw::cantilever::WingDesignConfig>()),
      data(std::make_shared<taw::cantilever::WingDesignData>()),
      reporter(rtIO) {
  /* Setup design modes */
  design_mode_runner["mode_0"] = [this]() { standard_design(); };

  /* Setup mass modes */
  mass_mode_runner["mode_0"] = [this]() { flops_mass(); };
  mass_mode_runner["mode_1"] = [this]() { chiozzotto_wer_mass(); };
}

void Wing::initialize() {
  myRuntimeInfo->out << "Initializing wing design ... [START]" << std::endl;
  /* Create Airfoil directory in project path */
  rtIO->createAirfoilDataDir();

  /* Read config */
  config->read(rtIO->moduleConfig);

  /* Read data */
  data->read(rtIO->acxml);

  /* fuselage must be read seperately since path to geometry needed */
  std::filesystem::path fuselage_path = rtIO->getGeometryDir();
  data->read_fuselage(rtIO->acxmlAccess, fuselage_path);

  /* Store method to variable */
  selected_design_mode = config->wing_design_mode.value();
  selected_mass_mode = config->wing_mass_mode.value();

  airfoils_library = std::make_shared<Airfoils>(static_cast<std::filesystem::path>(config->common_airfoil_data_path.value()));
  airfoils_library->add_directory_airfoils(rtIO->getAirfoilDataDir());

  myRuntimeInfo->out << "Initializing wing design ... [FINISHED]" << std::endl;
}
void Wing::run() {
  myRuntimeInfo->out << "Run wing design ... [START]" << std::endl;

  myRuntimeInfo->out << "Selected design method ... [START]" << std::endl;
  design_mode_runner[selected_design_mode]();
  myRuntimeInfo->out << "Selected design method ... [FINISHED]" << std::endl;

  myRuntimeInfo->out << "Selected mass method   ... [START]" << std::endl;
  mass_mode_runner[selected_mass_mode]();
  myRuntimeInfo->out << "Selected mass method   ... [FINISHED]" << std::endl;

  myRuntimeInfo->out << "Run wing design ... [FINISHED]" << std::endl;
}
void Wing::update() {
  myRuntimeInfo->out << "Update designed wing ... [START]" << std::endl;
  data->update(rtIO->acxml);
  myRuntimeInfo->out << "Update designed wing ... [FINISHED]" << std::endl;
}
void Wing::report() {
  myRuntimeInfo->out << "Report designed wing ... [START]" << std::endl;
  /* Sets body for html report */
  set_html_body();

  /* Generate reports */
  reporter.generateReports();
  /* Generate wing stl */
  geom2::Mesh wing_mesh = geom2::transform::to_mesh(data->wing);
  CGAL::IO::write_STL("wing_left.stl", wing_mesh);
  myRuntimeInfo->out << "Report designed wing ... [FINISHED]" << std::endl;
}
void Wing::save() {
  // nothing to do
}

/* == Design ========================================================================================================
 */

void Wing::standard_design() {
  /* Run standard method */
  myRuntimeInfo->out << "Wing Design ... [Start]" << std::endl;

  /* check if wing is kinked */
  bool is_wing_kinked = is_wing_kinked_check();

  /* get maximum thickness to chord ratio */
  double max_thickness_to_chord_ratio = get_max_thickness_to_chord_ratio();

  /* get mtom */
  double maximum_takeoff_mass = get_maximum_takeoff_mass();

  // Compute wing quarter chord based on selected mode
  const double installed_quarter_chord_sweep = get_wing_quarter_chord_sweep(
      config->sweep_calculation_mode.value(), config->user_sweep_angle_at_quarter_chord.value(),
      data->specification_design_mach.value(), data->sizing_point_wing_loading.value(),
      data->specification_design_altitude.value(), max_thickness_to_chord_ratio,
      config->delta_drag_divergence_to_mach_design.value(), config->korn_technology_factor.value());

  // Compute wing area based on selected mode
  const double installed_wing_area =
      get_design_wing_area(config->wing_area_calculation_mode.value(), config->user_wing_area.value(),
                           maximum_takeoff_mass, data->sizing_point_wing_loading.value());

  // Comptue aspect ratio and wing span
  std::map<std::string, std::tuple<double, double>> limits_field_length_and_wing_span =
      icao_easa_aerodrome_reference_code(data->specification_icao_aerodrome_reference_code.value());

  /* Current limits of wing span according to reference code data*/
  const auto [lower_wing_span, upper_wing_span] = limits_field_length_and_wing_span["wing_span"];

  /* Compute aspect ratio based on selected mode*/
  double aspect_ratio = get_design_aspect_ratio(config->aspect_ratio_calculation_mode.value(),
                                                config->user_aspect_ratio.value(), installed_quarter_chord_sweep);

  /* Determine span and possible aspect ratio limits */
  double span = sqrt(aspect_ratio * installed_wing_area);
  double maximum_possible_aspect_ratio = pow(upper_wing_span, 2.) / installed_wing_area;
  double minimum_possible_aspect_ratio = pow(lower_wing_span, 2.) / installed_wing_area;

  /* if current aspect ratio exceeds maximum possible aspect ratio -> set aspect ratio to maximum possible aspect ratio
   * and set span to according wing span */
  if (aspect_ratio > maximum_possible_aspect_ratio) {
    myRuntimeInfo->warn << "Calculated aspect ratio > maximum_possible_aspect_ratio" << std::endl;
    myRuntimeInfo->warn << "Switch to maximum possible aspect_ratio" << std::endl;
    aspect_ratio = maximum_possible_aspect_ratio;
    span = upper_wing_span;
  }
  // Check if current aspect ratio is below lower span limit -> set aspect ratio to minimum possible aspect ratio and
  // set span to according wing span */
  if (aspect_ratio < minimum_possible_aspect_ratio) {
    myRuntimeInfo->warn << "Current aspect ratio breach lower span limit ..." << std::endl;
    myRuntimeInfo->warn << "Switch to minimum possible aspect_ratio" << std::endl;
    aspect_ratio = minimum_possible_aspect_ratio;
    span = lower_wing_span;
  }

  /* Freeze aspect ratio and span */
  const double installed_aspect_ratio = aspect_ratio;
  const double installed_span = span;
  // Compute taper ratio based on selected mode
  const double installed_taper_ratio =
      get_design_taper_ratio(config->taper_ratio_calculation_mode.value(), config->user_taper_ratio.value(),
                             installed_aspect_ratio, installed_quarter_chord_sweep);

  // Comptue dihedral based on selected mode
  const double installed_dihedral = get_design_dihedral(
      config->dihedral_calculation_mode.value(), data->specification_wing_mounting.value(),
      config->user_dihedral.value(), installed_quarter_chord_sweep, data->specification_design_mach.value());

  // Generate wing
  if (is_wing_kinked) {
    data->wing =
        calculate_kinked_wing_geometry(installed_taper_ratio, installed_aspect_ratio, installed_quarter_chord_sweep,
                                       installed_dihedral, installed_wing_area, installed_span);
  } else {
    data->wing =
        calculate_unkinked_wing_geometry(installed_taper_ratio, installed_aspect_ratio, installed_quarter_chord_sweep,
                                         installed_dihedral, installed_wing_area, installed_span);
  }

  // Set wing name to constant name
  data->wing.name = "main_wing";

  data->wing_position.set_xyz(calculate_wing_position());

  // Set spar devices
  data->spars = get_spars(config->spars_mode.value(), config->user_defined_spars);

  // Set control devices
  double eta_kink = 0.;
  if (is_wing_kinked_check()) {
    eta_kink = get_design_relative_kink_position(
        config->relative_kink_position_calculation_mode.value(), config->user_relative_kink.value(),
        config->track_based_initial_relative_kink.value(), data->track_based_relative_kink.value());
  }
  auto [devices, deflections] = get_control_devices(config->control_device_mode.value(), config->user_defined_control_devices,
                                              data->spars, data->wing, data->fuselage, is_wing_kinked_check(), eta_kink,
                                              config->high_lift_device_type_leading_edge.value(),
                                              config->high_lift_device_type_trailing_edge.value());
  data->control_devices = devices;
  data->control_devices_deflections = deflections;

  /* Check control device boundaries - if inside fuselage */
  if ((config->control_device_mode.value() == "mode_0") &&
      (rtIO->aircraft_configuration_type() == "tube_and_wing")) {
    double eta_fuselage_max = geom2::measure::width_max(data->fuselage) / geom2::measure::span(data->wing);
    for (auto& device : data->control_devices) {
      size_t sec_id = 0;
      for (auto& section : device.sections) {
        if (section.origin.z() < eta_fuselage_max * 1.1) {
          myRuntimeInfo->warn << std::format(
              "User defined control device {} (section {}) inside fuselage width - set to span position to 1.1 "
              "fuselage width\n",
              device.name, sec_id);
          section.origin = geom2::Point_3(section.origin.x(), section.origin.y(), eta_fuselage_max * 1.1);
        }
        sec_id++;
      }
    }
  }
}

bool Wing::is_wing_kinked_check() {
  myRuntimeInfo->out << "Wing kink check ..." << std::endl;

  /* Check if mounting is low &  landing gear mounted on wing */
  if (data->specification_wing_mounting.value()== "low" &&
      data->specification_landing_gear_mounting.value() == "wing_mounted") {
    myRuntimeInfo->out << "Wing is kinked  ... [TRUE]" << std::endl;
    return true;
  } else {
    myRuntimeInfo->out << "Wing is kinked  ... [FALSE]" << std::endl;
    return false;
  }
}

double Wing::get_max_thickness_to_chord_ratio() {
  // Get maximum thickness to chord ratio based on wing profile and thickness mode
  // If user - get maximum thickness from elements
  // Else use max_thickness to chord ratio from sweep calculation method
  if (config->wing_profile_and_thickness_calculation_mode.value() == "mode_0") {
    std::vector<double> to_check;
    for (auto item : config->user_thickness_to_chord) {
      to_check.push_back(item.value());
    }
    return *std::max_element(to_check.begin(), to_check.end());
  } else {
    return config->max_thickness_to_chord_ratio.value();
  }
}

double Wing::get_maximum_takeoff_mass() {
  return data->mtom_mass_properties.data["mass"].value();
}

geom2::Point_3 Wing::calculate_wing_position() {
  // Get cg position - if overall not exists -> use fuselage cg
  double x_wing_apex, x_cg, y;
  std::tuple<double, double> z_limits;

  double y_mac_position = geom2::measure::mean_aerodynamic_chord_position(data->wing);
  double x_mac_position_quarter_chord_referred_to_centerline =
      geom2::measure::offset_LE(data->wing, y_mac_position).x() +
      0.25 * geom2::measure::chord(data->wing, y_mac_position);

  const double fuselage_length = geom2::measure::length(data->fuselage);
  const double fuselage_maximum_width = geom2::measure::width_max(data->fuselage);
  const double fuselage_maximum_height = geom2::measure::height_max(data->fuselage);

  double fuselage_upper_z;
  double fuselage_lower_z;

  /* Keel beam height as fraction of fuselage height*/
  double keel_beam_height = 0.005 * fuselage_maximum_height;  // Assumed 0.5% of fuselage height

  /* Define wing position limits */
  if (data->specification_wing_mounting.value() == "low") {
    fuselage_upper_z = data->fuselage.origin.z();
    fuselage_lower_z = data->fuselage.origin.z() - fuselage_maximum_height * 0.5;
    keel_beam_height = 0.005 * (fuselage_upper_z - fuselage_lower_z);
    z_limits = {fuselage_lower_z + keel_beam_height, fuselage_upper_z};
  } else if (data->specification_wing_mounting.value() == "high") {
    z_limits = {fuselage_lower_z, fuselage_upper_z - keel_beam_height};
  } else {
    z_limits = {fuselage_lower_z, fuselage_upper_z};
  }

  // Assume center position
  if (data->mtom_mass_properties.data["x"].value() < 0.01 && !data->fuselage.sections.empty()) {
    x_cg = fuselage_length * 0.45;
  } else {
    x_cg = data->mtom_mass_properties.data["x"].value();
  }
  auto const [lim_low, lim_up] = z_limits;
  double z_position = lim_low + 0.5 * (lim_up - lim_low);
  return geom2::Point_3(x_cg - x_mac_position_quarter_chord_referred_to_centerline, 0., z_position);
}

geom2::MultisectionSurface<geom2::AirfoilSection> Wing::calculate_kinked_wing_geometry(
    const double taper_ratio, const double aspect_ratio, const double quarter_chord_sweep, const double dihedral,
    const double wing_area, const double span) {
  double maximum_inner_trailing_edge_sweep = get_design_maximum_inner_trailing_edge_sweep(
      config->relative_kink_position_calculation_mode.value(),
      config->user_max_inner_trailing_edge_sweep.value(),          // old converted to radians
      config->track_based_max_inner_trailing_edge_sweep.value());  // old converted to radians

  const double max_thickness_to_chord_ratio = get_max_thickness_to_chord_ratio();

  // Get normalized half span values - eta
  const double half_span = 0.5 * span;

  const double half_fuselage_width = geom2::measure::width_max(data->fuselage) * 0.5;

  const double eta_fuselage = half_fuselage_width / half_span;
  double eta_centerline = 0. / half_span;
  double eta_tip = half_span / half_span;
  std::vector<double> eta_values = {eta_centerline, eta_fuselage, eta_tip};

  double eta_kink = get_design_relative_kink_position(
      config->relative_kink_position_calculation_mode.value(), config->user_relative_kink.value(),
      config->track_based_initial_relative_kink.value(), data->track_based_relative_kink.value());
  eta_values.push_back(eta_kink);

  std::vector<std::tuple<double, double, std::string>> wing_profile_and_thickness_vector =
      get_wing_profile_and_thickness(config->wing_profile_and_thickness_calculation_mode.value(),
                                     config->user_wing_profiles_and_thickness_vector,
                                     config->torenbeek_jenkinson_wing_profile.value(), max_thickness_to_chord_ratio,
                                     config->airfoil_critical_factor.value(), data->specification_design_mach.value(),
                                     data->specification_design_altitude.value(), quarter_chord_sweep,
                                     config->delta_drag_divergence_to_mach_design.value(),
                                     data->sizing_point_wing_loading.value(), eta_fuselage, eta_values);

  // Geometry computation functions
  auto condition_trailing_edge_max = [=](const double inner_taper_ratio) {
    return (4. / 3. * (tan(quarter_chord_sweep) - tan(maximum_inner_trailing_edge_sweep)) * half_span *
            (eta_kink - eta_fuselage) / (1 - inner_taper_ratio));
  };

  auto root_chord_function = [=](const double inner_taper_ratio) {
    return wing_area / (2. * half_span *
                        (eta_fuselage + (1 + inner_taper_ratio) * 0.5 * (eta_kink - eta_fuselage) +
                         (inner_taper_ratio + taper_ratio) * 0.5 * (eta_tip - eta_kink)));
  };

  // Maximum possible taper ratio
  const double eta_ratio = (eta_tip - eta_kink) / (eta_kink - eta_fuselage);
  const double taper_ratio_inner_max = (eta_ratio + taper_ratio) / (1 + eta_ratio);
  auto taper_ratio_inner_searcher = [=](const double inner_taper_ratio) {
    return root_chord_function(inner_taper_ratio) - condition_trailing_edge_max(inner_taper_ratio);
  };
  // Find root_chord which fulfills - half_wing_area = half_inner_wing_area + half_outer_wing_area +
  // half_fuselage_area
  double taper_ratio_inner = newton_algorithm(taper_ratio_inner_searcher, taper_ratio_inner_max);
  if (taper_ratio_inner > taper_ratio_inner_max) {
    taper_ratio_inner = taper_ratio_inner_max;
  }
  const double taper_ratio_outer = taper_ratio / taper_ratio_inner;
  const double root_chord = root_chord_function(taper_ratio_inner);

  // Compute values for wing geometry
  const double kink_chord = root_chord * taper_ratio_inner;
  const double tip_chord = root_chord * taper_ratio;
  const double fuselage_area = 2. * eta_fuselage * half_span * root_chord;
  const double inner_wing_area = (eta_kink - eta_fuselage) * (1 + taper_ratio_inner) * half_span * root_chord;
  const double outer_wing_area = (eta_tip - eta_kink) * (taper_ratio_inner + taper_ratio) * half_span * root_chord;

  // Inner wing
  const double tan_inner_wing_quarter_chord_sweep = tan(quarter_chord_sweep);
  const double tan_inner_wing_leading_edge_sweep =
      tan_inner_wing_quarter_chord_sweep +
      0.25 * root_chord / (half_span * (eta_kink - eta_fuselage)) * (1 - taper_ratio_inner);
  const double tan_inner_wing_trailing_edge_sweep =
      tan_inner_wing_quarter_chord_sweep -
      0.75 * root_chord / (half_span * (eta_kink - eta_fuselage)) * (1 - taper_ratio_inner);

  // Outer wing
  const double tan_outer_wing_quarter_chord_sweep = tan_inner_wing_quarter_chord_sweep;
  const double tan_outer_wing_leading_edge_sweep =
      tan_outer_wing_quarter_chord_sweep +
      0.25 * (root_chord * taper_ratio_inner) / (half_span * (eta_tip - eta_kink)) * (1 - taper_ratio_outer);
  const double tan_outer_wing_trailing_edge_sweep =
      tan_inner_wing_quarter_chord_sweep -
      0.75 * root_chord * taper_ratio_inner / (half_span * (eta_tip - eta_kink)) * (1 - taper_ratio_outer);

  // Check different conditions
  // Check sweeps leading edge and trailing edge by single equation
  // Condition 1: leading edge inner wing >= leading edge outer wing
  // Condition 2: trailing edge inner wing <= trailing edge outer wing
  // K_eta * (1-taper_ratio_2) >= taper_ratio_2 - taper_ratio -> OK
  const double K_eta = (eta_tip - eta_kink) / (eta_kink - eta_fuselage);
  std::cout << "Leading edge check ... ";
  if ((K_eta * (1 - taper_ratio_inner)) >= (taper_ratio_inner - taper_ratio)) {
    std::cout << "[OK]" << std::endl;
    // Check if trailing edge is none negative
    // s*(eta_kink-eta_fuselage)*tan(sweep_leading_edge_inner) >= 1 - taper_ratio_2 -> ok
    std::cout << "Trailing edge check ... ";
    if ((half_span * (eta_kink - eta_fuselage) * tan_inner_wing_leading_edge_sweep) >= (1 - taper_ratio_inner)) {
      std::cout << "[OK]" << std::endl;
    } else {
      throwError(__FILE__,__func__,__LINE__,"Error - trailing edge condition not fulfilled (TE Sweep < 0)");
    }
  } else {
    throwError(__FILE__,__func__,__LINE__,"Error - leading edge condition not fulfilled (LE Sweep inner < LE Sweep outer).");
  }

  /* Create Sections */
  std::vector<geom2::AirfoilSection> sections;
  geom2::MultisectionSurface<geom2::AirfoilSection> wing;
  wing.is_symmetric = true;
  geom2::Vector_3 section_position;
  for (auto profile_and_thickness : wing_profile_and_thickness_vector) {
    const auto [eta, t_to_c, profile] = profile_and_thickness;
    sections.push_back(get_rescaled_airfoil_geometry(profile_and_thickness));

    // Sections center -> fuselage
    if (eta <= eta_fuselage) {
      sections.back().origin = geom2::Point_3(0.0, 0.0, -eta * half_span);
      sections.back().set_chord_length(root_chord);
    }
    // Sections fuselage -> kink
    else if (eta > eta_fuselage && eta < eta_kink) {
      double normalized_eta = (eta - eta_fuselage) / (eta_kink - eta_fuselage);
      double x =
          std::lerp(0.0, tan_inner_wing_leading_edge_sweep * (eta_kink - eta_fuselage) * half_span, normalized_eta);
      double y = std::lerp(0.0, tan(dihedral) * (eta_kink - eta_fuselage) * half_span, normalized_eta);
      double z = -eta * half_span;
      double section_chord = std::lerp(root_chord, taper_ratio_inner * root_chord, normalized_eta);
      sections.back().origin = geom2::Point_3(x, y, z);
      sections.back().set_chord_length(section_chord);
    }
    // Sections kink -> tip
    else {
      double normalized_eta = (eta - eta_kink) / (eta_tip - eta_kink);
      double x = std::lerp(0.0, tan_outer_wing_leading_edge_sweep * (eta_tip - eta_kink) * half_span, normalized_eta) +
                 tan_inner_wing_leading_edge_sweep * (eta_kink - eta_fuselage) * half_span;
      double y = std::lerp(0.0, tan(dihedral) * (eta_tip - eta_kink) * half_span, normalized_eta) +
                 tan(dihedral) * (eta_kink - eta_fuselage) * half_span;
      double z = -eta * half_span;
      double section_chord = std::lerp(taper_ratio_inner * root_chord, taper_ratio * root_chord, normalized_eta);
      sections.back().origin = geom2::Point_3(x, y, z);
      sections.back().set_chord_length(section_chord);
    }
    sections.back().name = profile;
    wing.sections.push_back(sections.back());
    wing.normal = geom2::Direction_3(0., -1., 0.);
    // wing.origin = geom2::Point_3(0.45 * 37.5 - 0.5 * root_chord, 0., 0.);
    wing.origin = geom2::Point_3(0., 0., 0.);
  }
  // Create Sections

  return wing;
}

geom2::MultisectionSurface<geom2::AirfoilSection> Wing::calculate_unkinked_wing_geometry(
    const double taper_ratio, const double aspect_ratio, const double quarter_chord_sweep, const double dihedral,
    const double wing_area, const double span) {
  const double max_thickness_to_chord_ratio = get_max_thickness_to_chord_ratio();

  // Get normalized half span values - eta
  const double half_span = 0.5 * span;
  const double half_fuselage_width =
      geom2::measure::width(data->fuselage, geom2::measure::length(data->fuselage) * 0.5) * 0.5;
  const double eta_fuselage = half_fuselage_width / half_span;
  double eta_centerline = 0. / half_span;
  double eta_tip = half_span / half_span;
  std::vector<double> eta_values = {eta_centerline, eta_fuselage, eta_tip};

  std::vector<std::tuple<double, double, std::string>> wing_profile_and_thickness_vector =
      get_wing_profile_and_thickness(config->wing_profile_and_thickness_calculation_mode.value(),
                                     config->user_wing_profiles_and_thickness_vector,
                                     config->torenbeek_jenkinson_wing_profile.value(), max_thickness_to_chord_ratio,
                                     config->airfoil_critical_factor.value(), data->specification_design_mach.value(),
                                     data->specification_design_altitude.value(), quarter_chord_sweep,
                                     config->delta_drag_divergence_to_mach_design.value(),
                                     data->sizing_point_wing_loading.value(), eta_fuselage, eta_values);

  // Geometry computation functions

  const double root_chord = wing_area / (half_span * (2 * eta_fuselage + (1 + taper_ratio) * (eta_tip - eta_fuselage)));
  // Compute values for wing geometry

  // Outer wing
  const double tan_outer_wing_quarter_chord_sweep = tan(quarter_chord_sweep);
  const double tan_outer_wing_leading_edge_sweep =
      tan_outer_wing_quarter_chord_sweep +
      0.25 * root_chord / (half_span * (eta_tip - eta_fuselage)) * (1 - taper_ratio);

  std::vector<geom2::AirfoilSection> sections;
  geom2::MultisectionSurface<geom2::AirfoilSection> wing;
  wing.is_symmetric = true;
  geom2::Vector_3 section_position;
  for (auto profile_and_thickness : wing_profile_and_thickness_vector) {
    const auto [eta, t_to_c, profile] = profile_and_thickness;
    sections.push_back(get_rescaled_airfoil_geometry(profile_and_thickness));

    // Sections center -> fuselage
    if (eta <= eta_fuselage) {
      sections.back().origin = geom2::Point_3(0.0, 0.0, -eta * half_span);
      sections.back().set_chord_length(root_chord);
    }
    // Sections fuselage -> tip
    else {
      double normalized_eta = (eta - eta_fuselage) / (eta_tip - eta_fuselage);
      double x =
          std::lerp(0.0, tan_outer_wing_leading_edge_sweep * (eta_tip - eta_fuselage) * half_span, normalized_eta);
      double y = std::lerp(0.0, tan(dihedral) * (eta_tip - eta_fuselage) * half_span, normalized_eta);
      double z = -eta * half_span;
      double section_chord = std::lerp(root_chord, taper_ratio * root_chord, normalized_eta);
      sections.back().origin = geom2::Point_3(x, y, z);
      sections.back().set_chord_length(section_chord);
    }

    sections.back().name = profile;
    wing.sections.push_back(sections.back());
    wing.normal = geom2::Direction_3(0., -1., 0.);
    wing.origin = geom2::Point_3(0., 0., 0.);
  }

  // Create Sections
  return wing;
}

void Wing::flops_mass() {
  myRuntimeInfo->out << "Wing mass estimation using ... FLOPS" << std::endl;
  myRuntimeInfo->out << "Calculating ... " << std::endl;
  auto get_max_thickness_to_chord_ratio = [](const geom2::AirfoilSection section) {
    std::vector<double> t_to_c;
    const double chord = section.get_chord_length();
    const int steps = 50;
    const double step_width = chord / steps;
    for (size_t i = 0; i < steps; ++i) {
      t_to_c.push_back(geom2::measure::thickness(section, i * step_width) / chord);
    }
    return *std::max_element(t_to_c.begin(), t_to_c.end());
  };

  myRuntimeInfo->warn << "Following values are Hard coded!" << std::endl;
  myRuntimeInfo->warn << "================================" << std::endl;
  myRuntimeInfo->warn << "\twing_mounted_engines" << std::endl;
  myRuntimeInfo->warn << "================================" << std::endl;
  const double to_foot = convertUnit(METER, FOOT, 1.0);
  const double to_meter = 1. / to_foot;
  const double taper_ratio = geom2::measure::taper_ratio(data->wing);
  const double span_ft = geom2::measure::span(data->wing) * to_foot;
  const double wing_area_sqrft = geom2::measure::reference_area(data->wing) * to_foot * to_foot;
  const double aspect_ratio = geom2::measure::aspect_ratio(data->wing);
  const double t_to_c_root = get_max_thickness_to_chord_ratio(data->wing.sections.front());
  const double t_to_c_tip = get_max_thickness_to_chord_ratio(data->wing.sections.back());
  const double quarter_chord_sweep =
      geom2::measure::sweep(data->wing, 0.25 * span_ft * to_meter, 0.25);  // measure at 50% half span -> 25% of span

  const double to_lb = convertUnit(KILO, GRAM, NOPREFIX, POUND, 1.0);
  const double to_kg = 1 / to_lb;

  const double dg = data->mtom_mass_properties.data["mass"].value() * to_lb;

  // Tangent of 3/4 chord wing sweep
  double TLAM = tan(quarter_chord_sweep) - 2.0 * (1.0 - taper_ratio) / (aspect_ratio * (1.0 + taper_ratio));
  // Sine of 3/4 chord wing sweep angle
  double SLAM = TLAM / (sqrt(1 + TLAM * TLAM));

  // Factor Eq. 15
  double C4 = 1.0 - 0.5 * config->flops_faert.value();

  // Factor Eq. 16
  double C6 = 0.5 * config->flops_faert.value() - 0.16 * config->flops_fstrt.value();

  // Aspect ratio defined factor Eq. 17
  double CAYA = aspect_ratio > 5. ? aspect_ratio - 5.0 : 0.;
  // Wing sweep and aeroelastic factor
  double CAYL = (1.0 - SLAM * SLAM) * (1.0 + C6 * SLAM * SLAM + 0.03 * CAYA * C4 * SLAM);

  // Wing strut bracing factor (for cantilever w/o wing strut)
  double EMS = 1.0 - 0.25 * config->flops_fstrt.value();
  // Weighted average of the wing thickness to chord ratio (Jenkinson 1999 formulation)
  double TCA = 0.25 * (3. * t_to_c_tip + t_to_c_root);
  // wing equivalent bending material factor Dimension [-]
  double BT = 0.215 * (0.37 + 0.7 * taper_ratio) * pow(span_ft * span_ft / wing_area_sqrft, EMS) / (CAYL * TCA);

  // Total wing bending material weight

  // A1 - Table 1 Flops wing weight constants for transport and hwb
  const double A1 = 8.80;

  // A2
  const double A2 = 6.25;

  // ULF - Structural Ultimate load factor - default 3.75
  const double ULF = 3.75;

  // CAYF - only single fuselage usage
  const double CAYF = 1.0;

  // VARSWP - no variable sweep allowed
  const double VARSWP = 0.0;
  // VFACT
  const double VFACT = 1.0;
  // PCTL
  const double PCTL = 1.0;

  // Wing bending material weight, not including the effects of inertia [lb]
  const double W1NIR = A1 * BT * (1.0 + sqrt(A2 / span_ft)) * ULF * span_ft *
                       (1.0 - 0.4 * config->flops_fcomp.value()) * (1.0 - 0.1 * config->flops_faert.value()) * CAYF *
                       VFACT * PCTL / 1000000.0;

  // Total wing shear material and control surface weight
  const double A3 = 0.68;
  const double A4 = 0.34;
  const double A5 = 0.6;

  // total movable wing surface area including flaps, elevators, spoilers etc [ft^2]

  const double SFLAP = get_control_device_area_in_fraction_of_wing_area() *
                       wing_area_sqrft;  // set to 20% of wing_area - testing only TODO

  const double W2 = A3 * (1.0 - 0.17 * config->flops_fcomp.value()) * pow(SFLAP, A4) * pow(dg, A5);

  // Total wing miscellaneous items weight
  const double A6 = 0.035;
  const double A7 = 1.50;

  const double W3 = A6 * (1.0 - 0.3 * config->flops_fcomp.value()) * pow(wing_area_sqrft, A7);

  // Wing bending material weight intertia relief adjustment
  const int NEW = data->specification_number_of_wing_mounted_engines;
  const double CAYE = 1.0 - 0.03 * NEW;

  const double W1 = (dg * CAYE * W1NIR + W2 + W3) / (1.0 + W1NIR) - W2 - W3;

  // Total aft body weight for hybrid wing body aircraft - here no HWB
  const double W4 = 0.0;

  // Total wing weight
  const double wing_mass = (W1 + W2 + W3 + W4) * to_kg;

  myRuntimeInfo->out << "Calculated wing mass ... " << wing_mass << "[kg]" << std::endl;
  myRuntimeInfo->out << "Wing bending material mass ... " << W1 * to_kg << "[kg]" << std::endl;
  myRuntimeInfo->out << "Wing shear material and control surface mass ... " << W2 * to_kg << "[kg]" << std::endl;
  myRuntimeInfo->out << "Wing miscellaneous items mass ... " << W3 * to_kg << "[kg]" << std::endl;
  data->wing_mass_properties.data["mass"] = wing_mass;

  // Calculate wing center of gravity according to Howe 2005 page 252
  myRuntimeInfo->out << "Calculate Center of Gravity ..." << std::endl;
  const double half_span_mac_position = geom2::measure::mean_aerodynamic_chord_position(data->wing);
  const double mean_chord = geom2::measure::reference_area(data->wing) / geom2::measure::span(data->wing);
  const double mac = geom2::measure::chord(data->wing, half_span_mac_position);
  const geom2::Point_3 leading_edge_point = geom2::measure::offset_LE(data->wing, half_span_mac_position);
  data->wing_mass_properties.data["x"].set_value(data->wing_position.x.value() + leading_edge_point.x() + 0.25 * mac +
                                                 0.1 * mean_chord);
  // Transform values into global coordinate system z-> -y z
  data->wing_mass_properties.data["y"].set_value(data->wing_position.y.value() + fabs(leading_edge_point.z()) -
                                                 fabs(leading_edge_point.z()));
  data->wing_mass_properties.data["z"].set_value(data->wing_position.z.value() + leading_edge_point.y());

  myRuntimeInfo->out << "CG Position..." << std::endl;
  myRuntimeInfo->out << "x ... " << data->wing_mass_properties.data["x"].value() << std::endl;
  myRuntimeInfo->out << "y ... " << data->wing_mass_properties.data["y"].value() << std::endl;
  myRuntimeInfo->out << "z ... " << data->wing_mass_properties.data["z"].value() << std::endl;

}

/**
 * \brief Chiozzotto weight estimation relationship mass estimation - assumes symmetric engine distribution on wings
 *
 */
void Wing::chiozzotto_wer_mass() {
  myRuntimeInfo->out << "Wing mass estimation using ... CHIOZZOTTO" << std::endl;
  myRuntimeInfo->out << "Calculating ... " << std::endl;


  std::map<std::string, std::tuple<double, double>> min_max_mtom = {
      {"AL ASW", {20000, 250000}},   {"AL USW", {20000, 250000}},   {"AL FSW", {20000, 250000}},
      {"CFRP ASW", {20000, 250000}}, {"CFRP USW", {20000, 250000}}, {"CFRP FSW", {20000, 250000}}};
  std::map<std::string, std::tuple<double, double>> min_max_w_to_s = {
      {"AL ASW", {3000, 8000}},   {"AL USW", {3000, 8000}},   {"AL FSW", {3000, 8000}},
      {"CFRP ASW", {3000, 8000}}, {"CFRP USW", {3000, 8000}}, {"CFRP FSW", {3000, 8000}}};
  std::map<std::string, std::tuple<double, double>> min_max_aspect_ratio = {
      {"AL ASW", {10, 20}},   {"AL USW", {10, 20}},   {"AL FSW", {8, 16}},
      {"CFRP ASW", {10, 20}}, {"CFRP USW", {10, 20}}, {"CFRP FSW", {8, 16}}};
  std::map<std::string, std::tuple<double, double>> min_max_sweep = {{"AL ASW", {10, 40}},  {"AL USW", {0, 10}},
                                                                     {"AL FSW", {-5, -25}}, {"CFRP ASW", {10, 40}},
                                                                     {"CFRP USW", {0, 10}}, {"CFRP FSW", {-5, -25}}};
  std::map<std::string, std::tuple<double, double>> min_max_t_to_c = {
      {"AL ASW", {0.08, 0.14}},   {"AL USW", {0.08, 0.18}},   {"AL FSW", {0.08, 0.16}},
      {"CFRP ASW", {0.08, 0.14}}, {"CFRP USW", {0.08, 0.18}}, {"CFRP FSW", {0.08, 0.16}}};
  std::map<std::string, std::tuple<double, double>> min_max_v_eas = {
      {"AL ASW", {130, 200}},   {"AL USW", {130, 200}},   {"AL FSW", {130, 200}},
      {"CFRP ASW", {130, 200}}, {"CFRP USW", {130, 200}}, {"CFRP FSW", {130, 200}}};
  std::map<std::string, std::tuple<double, double>> min_max_taper_ratio = {
      {"AL ASW", {0.1, 0.5}},   {"AL USW", {0.1, 0.5}},   {"AL FSW", {0.1, 0.5}},
      {"CFRP ASW", {0.1, 0.5}}, {"CFRP USW", {0.1, 0.5}}, {"CFRP FSW", {0.1, 0.5}}};

  /**< Parameter containing a table of the constants and exponents of conventional Aircraft Covers*/
  std::map<std::string, std::tuple<double, double, double, double, double, double, double, double>>
      lookup_table_covers = {{"AL ASW", {2.14e-3, 1.345, -0.623, 1.286, -1.601, -0.830, -0.010, 0.104}},
                             {"AL USW", {2.66e-3, 1.320, -0.731, 1.370, 6.733, -0.945, 0.110, 0.305}},
                             {"AL FSW", {2.82e-5, 1.373, -1.152, 2.006, -5.507, -1.549, 1.036, 0.698}},
                             {"CFRP ASW", {9.47e-4, 1.440, -0.699, 1.233, -1.044, -0.808, 0.017, 0.518}},
                             {"CFRP USW", {1.29e-3, 1.396, -0.929, 1.4481, 0.667, -1.051, 0.266, 0.633}},
                             {"CFRP FSW", {3.92e-5, 1.401, -1.111, 1.874, -4.084, -1.436, 0.889, 0.751}}};

  /**< Parameter containing a table of constants and exponents of conventional Aircraft spars and ribs*/
  std::map<std::string, std::tuple<double, double, double, double, double, double, double, double>>
      lookup_table_spars_and_ribs = {{"AL ASW", {1.86, 1.384, -0.972, -0.072, -0.072, 0.486, 0.043, -0.126}},
                                     {"AL USW", {2.99, 1.376, -1.090, 0.008, 5.376, 0.423, 0.112, -0.070}},
                                     {"AL FSW", {9.82e-2, 1.362, -1.319, 0.598, -3.711, -0.126, 0.673, 0.391}},
                                     {"CFRP ASW", {1.50e-1, 1.411, -0.832, 0.105, -0.235, 0.301, 0.057, 0.101}},
                                     {"CFRP USW", {2.85e-1, 1.391, -0.994, 0.210, 5.854, 0.194, 0.164, 0.189}},
                                     {"CFRP FSW", {2.31e-2, 1.370, -1.174, 0.650, -2.750, -0.204, 0.620, 0.454}}};

  /**< Parameter containing a table of constants  and exponents of conventional Aircraft engine relief*/
  std::map<std::string, std::tuple<double, double, double>> lookup_table_engine_relief = {
      {"AL ASW", {8.41, 3.017, 0.817}},   {"AL USW", {12.5, 2.177, 1.011}},   {"AL FSW", {25.4, 1.941, 1.204}},
      {"CFRP ASW", {15.3, 3.104, 0.909}}, {"CFRP USW", {13.6, 2.569, 0.910}}, {"CFRP FSW", {24.7, 2.117, 1.148}}};

  // Chiozzotto Engine relief factor k_e (5.5 - Page 116)
  // IF NUMBER OF ENGINES ON WING > 2 ->
  // double engine_relative_spanwise_position = 0.3;  // BOOKMARK
  // Vector with data for one half ->
  // More than one engine -> k_e = k_e_1 * k_e_2

  double wing_loading = data->sizing_point_wing_loading.value();
  if (data->sizing_point_wing_loading.unit() == "kg/m^2") {
    wing_loading *= G_FORCE;
  }

  atmosphere isa;
  const double mtom = data->mtom_mass_properties.data["mass"].value();
  const double aspect_ratio = geom2::measure::aspect_ratio(data->wing);
  const double taper_ratio = geom2::measure::taper_ratio(data->wing);
  const double half_span = geom2::measure::span(data->wing) * 0.5;
  const double speed_of_sound_at_zero = isa.getSpeedOfSoundISA(0.0);
  const double pressure_at_zero = isa.getPressureISA(0.0);
  const double pressure_at_design_altitude = isa.getPressureISA(data->specification_design_altitude.value());

  std::vector<std::tuple<double, double>> engines_mass_and_eta = data->wing_mounted_engine_data;

  // Calculation of EAS Design
  const double v_eas_design = data->specification_design_mach.value() * speed_of_sound_at_zero *
                              sqrt(pressure_at_design_altitude / pressure_at_zero);

  // No LRA sweep existent - use 1/4 chord sweep (deviation from chiozzotto P.112 -> referred to figure 5.3)
  const double quarter_chord_sweep = -geom2::measure::sweep(data->wing, 0.5 * half_span, 0.25);

  // Get material selection
  std::string material = config->chiozzotto_material.value() + " ";
  if (quarter_chord_sweep > ACCURACY_LOW) {
    material += "ASW";
  } else if (quarter_chord_sweep < -ACCURACY_LOW) {
    material += "FSW";
  } else {
    material += "USW";
  }
  myRuntimeInfo->out << "Wing material used ... " << material << std::endl;

  // Get design relative kink -
  double t_to_c_kink = 0.0;
  double eta_kink = 0.35;
  if (is_wing_kinked_check()) {
    eta_kink = get_design_relative_kink_position(
        config->relative_kink_position_calculation_mode.value(), config->user_relative_kink.value(),
        config->track_based_initial_relative_kink.value(), data->track_based_relative_kink.value());
  }

  // Find thickness to chord ratio
  const double chord_at_kink = geom2::measure::chord(data->wing, eta_kink * half_span);

  double eta_position_last = 0.;
  double max_thickness_last = 0.0;
  double eta_position_next = 0.;
  double max_thickness_next = 0.0;
  for (size_t idx = 0; idx < data->wing.sections.size() - 1; ++idx) {
    eta_position_last = fabs(data->wing.sections.at(idx).origin.z()) / half_span;
    eta_position_next = fabs(data->wing.sections.at(idx + 1).origin.z()) / half_span;
    // If kink at a specified section
    if (fabs(eta_position_last - eta_kink) < ACCURACY_MEDIUM) {
      t_to_c_kink = geom2::measure::thickness_max(data->wing.sections.at(idx)) / chord_at_kink;
    } else if (fabs(eta_position_next - eta_kink) < ACCURACY_MEDIUM) {
      t_to_c_kink = geom2::measure::thickness_max(data->wing.sections.at(idx)) / chord_at_kink;
    } else if (eta_position_last > eta_kink && eta_position_next < eta_kink) {
      double max_thickness_last = geom2::measure::thickness_max(data->wing.sections.at(idx));
      double max_thickness_next = geom2::measure::thickness_max(data->wing.sections.at(idx));
      double dimensionless_position = (eta_kink - eta_position_last) / (eta_position_last - eta_position_next);
      t_to_c_kink = std::lerp(max_thickness_last, max_thickness_next, dimensionless_position) / chord_at_kink;
    } else {
      // Do nothing
    }
  }

  // Check inputs
  if (!(chiozzotto_wer_check_limits(min_max_mtom.at(material), mtom))) {
    myRuntimeInfo->warn << "Chiozzotto limit check out of bounds ... mtom ... " << mtom << std::endl;
  }
  if (!(chiozzotto_wer_check_limits(min_max_w_to_s.at(material), wing_loading))) {
    myRuntimeInfo->warn << "Chiozzotto limit check out of bounds ... wing_loading ... " << wing_loading << std::endl;
  }
  if (!(chiozzotto_wer_check_limits(min_max_aspect_ratio.at(material), aspect_ratio))) {
    myRuntimeInfo->warn << "Chiozzotto limit check out of bounds ... aspect ratio ... " << aspect_ratio << std::endl;
  }
  if (!(chiozzotto_wer_check_limits(min_max_taper_ratio.at(material), taper_ratio))) {
    myRuntimeInfo->warn << "Chiozzotto limit check out of bounds ... taper ratio ... " << taper_ratio << std::endl;
  }
  if (!(chiozzotto_wer_check_limits(min_max_sweep.at(material), quarter_chord_sweep * geom2::detail::to_degrees))) {
    myRuntimeInfo->warn << "Chiozzotto limit check out of bounds ... sweep ... "
                        << quarter_chord_sweep * geom2::detail::to_degrees << std::endl;
  }
  if (!(chiozzotto_wer_check_limits(min_max_t_to_c.at(material), t_to_c_kink))) {
    myRuntimeInfo->warn << "Chiozzotto limit check out of bounds ... thickness to chord ... " << t_to_c_kink
                        << std::endl;
  }
  if (!(chiozzotto_wer_check_limits(min_max_v_eas.at(material), v_eas_design))) {
    myRuntimeInfo->warn << "Chiozzotto limit check out of bounds ... equivalent airspeed at design point ... "
                        << v_eas_design << std::endl;
  }

  const auto [C_er, E_eta_er, E_K_er] = lookup_table_engine_relief[material];

  // Engine relief factor
  double k_e = 1.;
  for (auto engine : engines_mass_and_eta) {
    const auto [engine_mass, engine_spanwise_position] = engine;
    /* If engines exist, the spanwise position is absolute, if engines not exist, the default values are in rel. half span */
    double engine_relative_spanwise_position = data->engines_exist ? std::fabs(engine_spanwise_position)/half_span : engine_spanwise_position;
    k_e *= (1 - C_er * pow(engine_relative_spanwise_position, E_eta_er) * pow(engine_mass / mtom, E_K_er));
  }
  // Secondary structure mass
  double mass_secondary_structure = 0.0443 * mtom * config->chiozzotto_technology_factor.value();

  // Mass of covers
  const auto [C_cov, mtom_cov, w_to_s_cov, ar_cov, cos_lambda_cov, t_to_c_kink_cov, v_eas_design_cov,
              one_plus_taper_cov] = lookup_table_covers[material];
  double mass_covers = (k_e * C_cov * pow(mtom, mtom_cov) * pow(wing_loading, w_to_s_cov) * pow(aspect_ratio, ar_cov) *
                        pow(cos(quarter_chord_sweep), cos_lambda_cov) * pow(t_to_c_kink, t_to_c_kink_cov) *
                        pow(v_eas_design, v_eas_design_cov) * pow(1 + taper_ratio, one_plus_taper_cov)) *
                       config->chiozzotto_technology_factor.value();

  // Mass of spars and ribs
  const auto [C_sr, mtom_sr, w_to_s_sr, ar_sr, cos_lambda_sr, t_to_c_kink_sr, v_eas_design_sr, one_plus_taper_sr] =
      lookup_table_spars_and_ribs[material];
  double mass_spars_and_ribs = (C_sr * pow(mtom, mtom_sr) * pow(wing_loading, w_to_s_sr) * pow(aspect_ratio, ar_sr) *
                                pow(cos(quarter_chord_sweep), cos_lambda_sr) * pow(t_to_c_kink, t_to_c_kink_sr) *
                                pow(v_eas_design, v_eas_design_sr) * pow(1 + taper_ratio, one_plus_taper_sr)) *
                               config->chiozzotto_technology_factor.value();

  double wing_mass = mass_covers + mass_spars_and_ribs + mass_secondary_structure;

  myRuntimeInfo->out << "Calculated wing mass ... " << wing_mass << "[kg]" << std::endl;
  myRuntimeInfo->out << "Wing mass spars and ribs ..." << mass_spars_and_ribs << "[kg]" << std::endl;
  myRuntimeInfo->out << "Wing mass covers  ..." << mass_covers << "[kg]" << std::endl;
  myRuntimeInfo->out << "Wing mass secondary structure ..." << mass_secondary_structure << "[kg]" << std::endl;
  data->wing_mass_properties.data["mass"] = wing_mass;

  // Calculate wing center of gravity according to Howe 2005 page 252
  myRuntimeInfo->out << "Calculate Center of Gravity ..." << std::endl;
  const double half_span_mac_position = geom2::measure::mean_aerodynamic_chord_position(data->wing);
  const double mean_chord = geom2::measure::reference_area(data->wing) / geom2::measure::span(data->wing);
  const double mac = geom2::measure::chord(data->wing, half_span_mac_position);
  const geom2::Point_3 leading_edge_point = geom2::measure::offset_LE(data->wing, half_span_mac_position);
  data->wing_mass_properties.data["x"].set_value(data->wing_position.x.value() + leading_edge_point.x() + 0.25 * mac +
                                                 0.1 * mean_chord);
  // Transform values into global coordinate system z-> -y z
  data->wing_mass_properties.data["y"].set_value(data->wing_position.y.value() + fabs(leading_edge_point.z()) -
                                                 fabs(leading_edge_point.z()));
  data->wing_mass_properties.data["z"].set_value(data->wing_position.z.value() + leading_edge_point.y());

  myRuntimeInfo->out << "CG Position..." << std::endl;
  myRuntimeInfo->out << "x ... " << data->wing_mass_properties.data["x"].value() << std::endl;
  myRuntimeInfo->out << "y ... " << data->wing_mass_properties.data["y"].value() << std::endl;
  myRuntimeInfo->out << "z ... " << data->wing_mass_properties.data["z"].value() << std::endl;
}

bool Wing::chiozzotto_wer_check_limits(std::tuple<double, double>& min_max, double value_to_check) {
  const auto [min, max] = min_max;
  if ((value_to_check > min && value_to_check < max) ||
      (fabs(value_to_check - min) < ACCURACY_MEDIUM || fabs(value_to_check - max) < ACCURACY_MEDIUM)) {
    return true;
  } else {
    return false;
  }
}

double Wing::get_control_device_area_in_fraction_of_wing_area() {
  double overall_control_device_area = 0.0;
  double wing_area = geom2::measure::reference_area(data->wing);
  for (auto& item : data->control_devices) {
    auto cd = geom2::transform::to_absolute(item, data->wing);
    overall_control_device_area += geom2::measure::area(cd);
  }
  return overall_control_device_area / wing_area;
}
  geom2::AirfoilSection Wing::get_rescaled_airfoil_geometry(std::tuple<double, double, std::string> wing_profile_and_thickness) {
  const auto [eta, t_to_c, wing_profile] = wing_profile_and_thickness;
  geom2::Polygon_2 shape = airfoils_library->get_airfoil(wing_profile);
  airfoils_library->copy_available_airfoil(wing_profile,rtIO->getAirfoilDataDir());
  geom2::AirfoilSection section(shape);
  double max_thickness_to_chord = geom2::measure::thickness_max(section) / section.get_chord_length();
  section.scale_thickness(t_to_c / max_thickness_to_chord);

  return section;
}
}  // namespace cantilever
}  // namespace taw

// Helper methods

